// Define the configuraton for the I/O
var DIGITAL_OUTPUTS_LOCATION = '/sys/class/gpio/gpio';
var DIGITAL_OUTPUTS_BASEGPIO = 496;
var DIGITAL_OUTPUTS_POINTS = 16;// Default = 16
var digitalOutputsArray = [
    { "Address" : "0", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "1", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "2", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "3", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "4", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 },
    { "Address" : "5", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "6", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "7", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 

    { "Address" : "8", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "9", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "10", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "11", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "12", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "13", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "14", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }, 
    { "Address" : "15", "Name" : "Name", "Value": 0 , "Enabled": "FALSE", "SendEvents": "FALSE", "Polarity": "Normally Open", "ActionType": 0 , "TimeChanged": 0 }
];

// Requires
var fs = require('fs');
var moment = require('moment');
var SmtpManager = [];

//#region Digital Outputs - Initialization
// Initialize the analog outputs upon system start
//      - Checks if database table exists, if not creates the table and fills it with default data for each point
//      - Else reads from the database and updates digitalOutputsArray
exports.InitializeDigitalOutputs = function (smtp, db, io, logger, callback) {
    // Initialize SmtpManager
    SmtpManager = smtp;
    
    // Try getting the DigitalOutputs table
    db.get("SELECT name FROM sqlite_master WHERE type='table' AND name='DigitalOutputs'", function (err, rows) {
        if (err !== null) {
            logger.error('Digital Outputs - Initialize Error (Check)=' + JSON.stringify(err));
        }
        else if (typeof rows === "undefined") {
            // We read nothing, so let's create the table ...
            logger.info("Digital Outputs - Initialize by creating database table ...");
            db.exec("CREATE TABLE if not exists DigitalOutputs (Address INTEGER, Name TEXT, Value TEXT, Enabled TEXT, SendEvents TEXT, Polarity TEXT, ActionType TEXT, TimeChanged TEXT )", function (err) {
                if (err !== null) {
                    logger.error('Digital Outputs - Initialize Error (Create)=' + JSON.stringify(err));
                }
                else {
                    var timeChanged = moment();// Record the create time as the last time the I/O point value was changed
                    
                    // Created the table, now add data ...
                    db.run("BEGIN")
                    for (var i = 0; i < DIGITAL_OUTPUTS_POINTS; i++) {
                        // Initialize the array
                        UpdateDigitalOutputConfiguration(i, "Digital Output " + (i + 1), 1, "FALSE", "FALSE", "Normally Open", "Latched", timeChanged, io, logger);
                        // Empty space here to be consistent
                        // Empty space here to be consistent
                        
                        // Initialize the database
                        insertStatement = i + ", 'Digital Output " + (i + 1) + "', '1', 'FALSE', 'FALSE', 'Normally Open', 'Latched', '" + timeChanged + "'";
                        db.run("INSERT INTO DigitalOutputs (Address, Name, Value, Enabled, SendEvents, Polarity, ActionType, TimeChanged) VALUES (" + insertStatement + ")", function (err) {
                            if (err !== null) {
                                logger.error('Digital Outputs - Initialize Error (Insert)=' + JSON.stringify(err));
                            }
                        });
                    }
                    db.run("COMMIT");
                    
                    // Digital Outputs table created
                    if (typeof callback == "function") {
                        logger.info('Digital Outputs - Initialized new database table');
                        callback();
                    }
                }
            });
        }
        else {
            // Read from the database and populate the digitalOutputsArray
            logger.info("Digital Outputs - Initialize by reading database ...");
            db.each("SELECT rowid as id, Address, Name, Value, Enabled, SendEvents, Polarity, ActionType, TimeChanged FROM DigitalOutputs ORDER BY Address", function (err, row) {
                if (err !== null) {
                    logger.error('Digital Outputs - Initialize Error (Select)=' + JSON.stringify(err));
                }
                else {
                    UpdateDigitalOutputConfiguration(row.Address, row.Name, row.Value, row.Enabled, row.SendEvents, row.Polarity, row.ActionType, row.TimeChanged, io, logger);
                }
            }, function () {                
                // Digital Outputs read from table
                if (typeof callback == "function") {
                    logger.info('Digital Outputs - Initialized from existing database table');
                    callback();
                }            
            });
        }
    });
};

// Adjust the digitalOutputsArray
module.exports.UpdateDigitalOutputConfiguration = UpdateDigitalOutputConfiguration;
function UpdateDigitalOutputConfiguration(Address, Name, Value, Enabled, SendEvents, Polarity, ActionType, TimeChanged, io, logger) {
    digitalOutputsArray[Address].Name = Name;
    digitalOutputsArray[Address].Value = Value;
    digitalOutputsArray[Address].Enabled = Enabled;
    digitalOutputsArray[Address].SendEvents = SendEvents;
    digitalOutputsArray[Address].Polarity = Polarity;
    digitalOutputsArray[Address].ActionType = ActionType;
    digitalOutputsArray[Address].TimeChanged = TimeChanged;
    sendValueDigitalOutputs(io, logger);// Update the Digital Outputs configuration
};

// Reset the digital outputs on start-up to ensure a proper value
exports.ResetDigitalOutputs = function (db, logger, io, netSnmp, ledManager, callback) {
    // Get the stored Polarity from the database,
    //  then adjust the reset point based on the polarity        
    var timeChanged = moment();
    var digitalOutputs = [];
    db.all("SELECT Address, Name, Value, Enabled, SendEvents, Polarity, ActionType, TimeChanged FROM digitaloutputs ORDER BY Address", function (err, rows) {
        if (err) {
            throw logger.error('Digital Outputs - Reset Error (Select)=' + JSON.stringify(err));
        }

        // For each point, reset the value based on polarity setting 
        for (index = 0; index < rows.length; index++) {

            if (rows[index].Polarity === "Normally Closed") {
                polarityValue = "1";
            } else {
                polarityValue = "0";
            }

            // Reset the output point ...        
            fs.writeFile(DIGITAL_OUTPUTS_LOCATION + (DIGITAL_OUTPUTS_BASEGPIO + index) + "/value", polarityValue, function (err) {
                if (err) {
                    return logger.error('Digital Outputs - Reset Error (Write)=' + JSON.stringify(err));
                }
            });
        }

        // Digital Outputs reset is complete
        if (typeof callback == "function") {
            logger.debug('Digital Outputs - Reset to initial values');
            callback();
        }
    });
}
//#endregion Digital Outputs - Initialization

//#region Digital Outputs - Control
exports.SetDigitalOutput = function (outputPoint, setValue, logger, db, callback) {
    logger.info('Digital Outputs.SetDigitalOutput');

    // Get the digital output settings to determine if this is NC or NO
    db.get("SELECT Polarity FROM digitaloutputs WHERE Address='" + outputPoint + "'", function (err, row) {
        if (typeof row !== "undefined") {
            // Ensure we know the proper polarity
            polarity = row.Polarity;
        }
        logger.info("Digital Outputs.SetDigitalOutput: Digital Output #" + (parseInt(outputPoint)+1) + " Set Value=" + setValue + " Polarity=" + row.Polarity);
        
        // Determine the value to set
        var index = DIGITAL_OUTPUTS_BASEGPIO + parseInt(outputPoint);
        if (polarity === "Normally Open") {
            if (setValue == "0") {
                value = "0";
            }
            else {
                value = "1";
            }
        } 
        else {
            if (setValue == "0") {
                value = "1";
            }
            else {
                value = "0";
            }
        }
        
        // Set the value ...
        var index = DIGITAL_OUTPUTS_BASEGPIO + parseInt(outputPoint);
        var result;
        fs.writeFile("/sys/class/gpio/gpio" + index + "/value", value, function (err) {
            if (err) {
                // Record the error
                result = JSON.stringify(err);
                logger.error("Digital Outputs.SetDigitalOutput: Digital Output #" + (parseInt(outputPoint) + 1) + " Set Value=" + setValue + " Result=" + result);
                
                // Return with error
                if (typeof callback == "function") {
                    callback(result);
                }
                return;
            }
            
            // Record the result
            result = 'OK';
            logger.debug('Digital Outputs.SetDigitalOutput: Digital Output #' + (parseInt(outputPoint) + 1) + ' Set Value=' + setValue + ' Result=' + result);
            
            // Return with OK !
            if (typeof callback == "function") {
                callback(result);
            }
        });
    });
}

//#endregion Digital Outputs - Control

//#region Digital Outputs - Processing
// Intervalled processing of digital outputs
var processCounter = 0;
var processTotalTime = 0;
exports.ProcessDigitalOutputs = function (db, logger, io, netSnmp, ledManager) {
    timeStart = moment();
    
    var index = 0;// This is an index to digital I/O point - accounts for first 24 bits being at a higher location that the second 24 bits
    var fileIndex = 0;// This is an index to the GPIO location - it's the index + the GPIO offset
    var timeChanged = moment();
    for (index = 0; index < DIGITAL_OUTPUTS_POINTS; index++) {
        fileIndex = DIGITAL_OUTPUTS_BASEGPIO + index;
        
        // Only read value if the I/O point is enabled
        if (digitalOutputsArray[index].Enabled === "TRUE") {
            try {
                data = fs.readFileSync(DIGITAL_OUTPUTS_LOCATION + fileIndex + "/value");
                readValue = String(data).replace(/(\r\n|\n|\r)/gm, "");
                
                // Normally Closed vs Normally Open
                if (digitalOutputsArray[index].Polarity == "Normally Closed") {
                    if (readValue === "1") {
                        readValue = "0";
                    } else {
                        readValue = "1";
                    }
                }

                // Check to see if value has changed, if so update database and clients ...
                if (digitalOutputsArray[index].Value != readValue) {
                    digitalOutputsArray[index].Value = readValue;
                    digitalOutputsArray[index].TimeChanged = timeChanged;
                    sendSocketSnmpSmtp(digitalOutputsArray[index], db, io, logger, netSnmp)

                    // Value has changed, update the database and then the clients ...
                    db.run("UPDATE DigitalOutputs SET Value='" + readValue 
                        + "', TimeChanged='" + timeChanged 
                        + "' WHERE Address = '" + index + "'", 
                        function () {
                    });
                }
                //logger.info("Digital Output[" + (index+1) + "]=" + readValue + " (Enabled)")
            }
            catch (error) {
                logger.error('Digital Outputs - Process Error (Read)=' + JSON.stringify(error));
            }
        }
        else {
            //logger.info("Digital Output[" + (index+1) + "]=" + digitalOutputsArray[index].Value + " (Disabled)")
        }
    }
    //if (processCounter === 500) {
    //    logger.info("ProcessDigitalOutputs: Average process time = " + (processTotalTime / processCounter).toFixed(2) + "msec");
    //    processCounter = 0;
    //    processTotalTime = 0;
    //}
    //processCounter++;
    //processTotalTime += (Date.now() - timeStart);
}
//#endregion Digital Outputs - Processing

//#region Digital Outputs - Send Information
// Initialize the snmpHostSessionsArray used for send SNMP messages
var SnmpHostSessionsArray = [];// Instance array for list of SNMP host sessions
exports.SetSnmpHostSessionsArray = function (snmpHostSessions, logger,callback) {
    SnmpHostSessionsArray = snmpHostSessions.slice(0);// Copy array
    logger.debug("Digital Outputs - Initialized SNMP Host Sessions");
    
    if (typeof callback == "function") {
        callback();
    }
}

function sendSocketSnmpSmtp(digitalOutput, db, io, logger, netSnmp) {
    logger.info("Digital Outputs - [Value UPDATED] Digital Output[" + (parseInt(digitalOutput.Address) + 1) + "] '" + digitalOutput.Name + "' Value=" + digitalOutput.Value);
    logger.debug("Digital Outputs - [Value UPDATED] Digital Output[" + (parseInt(digitalOutput.Address) + 1) + "] =" + JSON.stringify(digitalOutput));
    
    // SOCKET.IO: Send changed value to all attached clients
    sendValueDigitalOutput("DigitalOutput", digitalOutput, io, logger);
    
    // SNMP: Send SNMP Trap/Notification if threshold(s) crossed
    sendSNMP(digitalOutput, io, logger, netSnmp);

    // SMTP: Send SMTP message if threshold(s) crossed
    sendSMTP(digitalOutput, io, logger)
}

function sendValueDigitalOutput(dataType, data, io, logger) {
    io.sockets.emit(dataType, data);
}

function sendValueDigitalOutputs(io, logger) {
    enabled = 0;
    for (i = 0; i < digitalOutputsArray.length; i++) {
        if (digitalOutputsArray[i].Enabled === "TRUE") enabled++;
    }
    io.sockets.emit("DigitalOutputs", { "DigitalOutputsTotal": digitalOutputsArray.length, "DigitalOutputsEnabled": enabled })
}

function sendSNMP(ioPoint, io, logger, netSnmp) {
    // SNMP definitions - also refer to the TASC-SUMMIT.MIB file for detailed information
    if (ioPoint.Enabled == "FALSE") return;
    if (ioPoint.SendEvents == "FALSE") return;
    
    address = parseInt(ioPoint.Address) + 1;// To overcome concatenation
    logger.info("Digital Outputs - Sending SNMP Trap for Digital Output[" + address + "] + '" + ioPoint.Name + "'");
    
    // Set up the SNMP trap    
    var trapOid = "1.3.6.1.4.1.11902.3.2.1000.1000.0." + (address - 1 + 1000);
    var varbinds = [
        {
            // Timestamp
            oid: "1.3.6.1.4.1.11902.3.2." + (address) + ".0.1",
            type: netSnmp.ObjectType.OctetString,
            value: new Buffer(moment().toString())
        },
        {
            // Address
            oid: "1.3.6.1.4.1.11902.3.2." + (address) + ".0.2",
            type: netSnmp.ObjectType.Integer,
            value: parseInt(ioPoint.Address)+1
        },
        {
            // Name
            oid: "1.3.6.1.4.1.11902.3.2." + (address) + ".0.3",
            type: netSnmp.ObjectType.OctetString,
            value: new Buffer(ioPoint.Name)
        },
        {
            // Value ("ON" or "OFF")
            oid: "1.3.6.1.4.1.11902.3.2." + (address) + ".0.4",
            type: netSnmp.ObjectType.OctetString,
            value: new Buffer((ioPoint.Value === "1") ? "ON":"OFF")
        },
        {
            // Status (1=ON, 0=OFF)
            oid: "1.3.6.1.4.1.11902.3.2." + (address) + ".0.5",
            type: netSnmp.ObjectType.Integer,
            value: (ioPoint.Value === "1") ? 1:0
        },
        {
            // Polarity (1=Normally Closed, 0=Normally Open)
            oid: "1.3.6.1.4.1.11902.3.2." + (address) + ".0.6",
            type: netSnmp.ObjectType.Integer,
            value: (ioPoint.Polarity === "Normally Closed") ? 1: 0
        }
    ];
    
    // Send the SNMP notification
    for (i = 0; i < SnmpHostSessionsArray.length; i++) {
        // Skip if this host is not enabled
        if (SnmpHostSessionsArray[i].Enabled === "FALSE") continue;
        
        // Send trap to host ...
        IPAddress = SnmpHostSessionsArray[i].Session.target;
        logger.debug("Digital Outputs - Sending to SNMP Host @ " + IPAddress + " - Digital Output[" + varbinds[1].value + "] '" + varbinds[2].value + "' Value=" + varbinds[4].value);
        SnmpHostSessionsArray[i].Session.trap(trapOid, varbinds, function (error) {
            if (error)
                logger.error("Digital Outputs - Send SNMP Error=" + error);
        });
    }
}

function sendSMTP(ioPoint, io, logger) {
    // SNMP definitions - also refer to the TASC-SUMMIT.MIB file for detailed information
    if (ioPoint.Enabled == "FALSE") return;
    if (ioPoint.SendEvents == "FALSE") return;
    
    // Set up the SMTP Email message
    var message = [];
    address = parseInt(ioPoint.Address) + 1;// To overcome concatenation
    var event = "'" + ioPoint.Name + "' has changed to status=" + ((ioPoint.Value === "1") ? "ON":"OFF");
    message += moment().toString() + "\n";
    message += "Address=" + (parseInt(ioPoint.Address) + 1) + "\n";
    message += "Name=" + ioPoint.Name + "\n";
    message += "Value=" + ((ioPoint.Value === "1") ? "ON":"OFF") + "\n";
    message += "Status=" + ((ioPoint.Value === "1") ? 1:0) + "\n";
    message += "Polarity=" + ((ioPoint.Polarity === "Normally Closed") ? 1: 0) + "\n";
    
    var ioInfo = "Digital Output[" + address + "] '" + ioPoint.Name + "'";
    
    logger.info("Digital Outputs - Sending SMTP Event for Digital Outputs[" + address + "] '" + ioPoint.Name + "' Value=" + ((ioPoint.Value === "1") ? "ON":"OFF"));
    SmtpManager.SendSmtp(message, ioInfo, event, function (result) {
        logger.trace("Digital Outputs - SMTP Email Event Response for Digital Output[" + address + "] '" + ioPoint.Name + "' Value=" + ((ioPoint.Value === "1") ? "ON":"OFF") + ": " + result.result);
    })
}

//#endregion Digital Outputs - Send Information